program scandatetest;
{
  Draft for an inverse of formatdatetime, destined for the dateutils unit of FPC.
  (C) 2007 Marco van de Voort
  License: modified BSD (no "prominent" notice required. Attribution in src is enough.)

  Limitations/implementation details:
  - An inverse of FormatDateTime is not 100% an inverse, simply because one can put e.g. time tokens twice in the format string,
       and scandatetime wouldn't know which time to pick.
  - Strings like hn can't be reversed safely. E.g. 1:2 (2 minutes after 1) delivers 12 which is parsed as 12:00 and then
       misses chars for the "n" part.
  - trailing characters are ignored.
  - no support for Eastern Asian formatting characters since they are windows only.
  - no MBCS support.

  Extensions
  - #9 eats whitespace.
  - whitespace at the end of a pattern is optional.
  - ? matches any char.
  - Quote the above chars to really match the char.
}

{$ifdef FPC}
  {$mode Delphi}
{$else}
  {$APPTYPE CONSOLE}
{$endif}


uses
  SysUtils,dateutils;

const whitespace  = [' ',#13,#10];
      hrfactor    = 1/(24);
      minfactor   = 1/(24*60);
      secfactor   = 1/(24*60*60);
      mssecfactor = 1/(24*60*60*1000);

const AMPMformatting : array[0..2] of string =('am/pm','a/p','ampm');

procedure raiseexception(const s:string);

begin
  raise EConvertError.Create(s);
end;

function scandatetime(const pattern:string;s:string;startpos:integer=1) : tdatetime;

var len ,ind  : integer;
    yy,mm,dd  : integer;
    timeval   : TDateTime;
    activequote: char;

procedure intscandate(ptrn:pchar;plen:integer);

var
    pind : integer;

function findimatch(const mnts:array of string;p:pchar):integer;
var i : integer;
begin
  result:=-1;
  i:=0;
  while (i<=high(mnts)) and (result=-1) do
    begin
      if AnsiStrLIComp(p,@mnts[i][1],length(mnts[i]))=0 then
        result:=i;
      inc(i);
    end;
end;

function findmatch(const mnts : array of string;const s:string):integer;

var i,j:integer;
begin
  result:=findimatch(mnts,@s[ind]);
  if result=-1 then
    raiseexception('Can''t find valid value at pattern position '+inttostr(pind)+' string position '+inttostr(ind))
  else
    begin
      inc(ind,length(mnts[result])+1);
      inc(pind,length(mnts[result])+1);
      inc(result); // was 0 based.
    end;
end;

var
    pivot,
    i,j,k     : integer;

function scanfixedint(maxv:integer):integer;
var c : char;
    oi:integer;
begin
  result:=0;
  oi:=ind;
  c:=ptrn[pind];
  while (pind<plen) and (ptrn[pind]=c) do inc(pind);
  while (maxv>0) and (ind<=len) and (s[ind] IN ['0'..'9']) do
    begin
      result:=result*10+ord(s[ind])-48;
      inc(ind);
      dec(maxv);
    end;
  if oi=ind then
    raiseexception('Pattern mismatch char "'+c+'" at position '+inttostr(oi));
end;

procedure matchchar(c:char);

begin
  if (ind>len) or (s[ind]<>c) then
    raiseexception('Pattern mismatch char "'+s[ind]+'" <> "'+ptrn[pind]+'" at position '+inttostr(ind));
  inc(pind);
  inc(ind);
end;

function scanpatlen:integer;
var c : char;
    lind : Integer;
begin
  result:=pind;
  lind:=pind;
  c:=ptrn[lind];

  while (lind<=plen) and (ptrn[lind]=c) do
      inc(lind);
  result:=lind-result;
end;

procedure matchpattern(const lptr:string); {$ifdef FPC}inline;{$endif}

var len:integer;
begin
  len:=length(lptr);
  if len>0 then
    intscandate(@lptr[1],len);
end;

var lasttoken,lch : char;

begin
  pind:=0;     lasttoken:=' ';
  while (ind<=len) and (pind<plen) do
     begin
       lch:=upcase(ptrn[pind]);
       if activequote=#0 then
          begin
            if (lch='M') and (lasttoken='H') then
              begin
                i:=scanpatlen;
                if i>2 then
                  raiseexception(' mm in a sequence hh:mm is interpreted as time(nn). No longer versions allowed!');
                timeval:=timeval+scanfixedint(2)* minfactor;
              end
            else
            case lch of
               'H':  timeval:=timeval+scanfixedint(2)* hrfactor;
               'D':  begin
                       i:=scanpatlen;
                       case i of
                          1,2 : dd:=scanfixedint(2);
                          3   : dd:=findmatch(shortDayNames,s);
                          4   : dd:=findmatch(longDayNames,s);
                          5   : matchpattern(shortdateformat);
                          6   : matchpattern(longdateformat);
                         end;
                     end;
               'N':  timeval:=timeval+scanfixedint(2)* minfactor;
               'S':  timeval:=timeval+scanfixedint(2)* secfactor;
               'Z':  timeval:=timeval+scanfixedint(3)* mssecfactor;
               'Y':  begin
                       i:=scanpatlen;
                       yy:=scanfixedint(i);
                       if i<=2 then
                         begin
                           pivot:=YearOf(now)-TwoDigitYearCenturyWindow;
                           inc(yy, pivot div 100 * 100);
                           if (TwoDigitYearCenturyWindow > 0) and (yy < pivot) then
                              inc(yy, 100);
                         end;
                      end;
               'M': begin
                       i:=scanpatlen;
                       case i of
                          1,2: mm:=scanfixedint(2);
                          3:   mm:=findmatch(ShortMonthNames,s);
                          4:   mm:=findmatch(LongMonthNames,s);
                          end;
                    end;
               'T' : begin
                       i:=scanpatlen;
                       case i of
                       1: matchpattern(shortdateformat);
                       2: matchpattern(longtimeformat);
                       end;
                     end;
               'A' : begin
                            i:=findimatch(AMPMformatting,@ptrn[pind]);
                            case i of
                              0: begin
                                   i:=findimatch(['AM','PM'],@s[ind]);
                                   case i of
                                     0: ;
                                     1: timeval:=timeval+12*hrfactor;
                                   else
                                     raiseexception('AM/PM mismatch');
                                     end;
                                   inc(pind,length(AMPMformatting[0]));
                                   inc(ind,2);
                                 end;
                              1: begin
                                    case upcase(s[ind]) of
                                     'A' : ;
                                     'P' : timeval:=timeval+12*hrfactor;
                                   else
                                     raiseexception('AM/PM mismatch');
                                     end;
                                   inc(pind,length(AMPMformatting[1]));
                                   inc(ind);
                                 end;
                               2: begin
                                    i:=findimatch([timeamstring,timepmstring],@s[ind]);
                                    case i of
                                     0: inc(ind,length(timeamstring));
                                     1: begin
                                          timeval:=timeval+12*hrfactor;
                                          inc(ind,length(timepmstring));
                                        end;
                                   else
                                     raiseexception('AM/PM mismatch');
                                     end;
                                   inc(pind,length(AMPMformatting[2]));
                                   inc(pind,2);
                                   inc(ind,2);
                                 end;
                            else  // no AM/PM match. Assume 'a' is simply a char
                                matchchar(ptrn[pind]);
                             end;
                         end;
               '/' : matchchar(DateSeparator);
               ':' : begin
                       matchchar(TimeSeparator);
                       lch:=lasttoken;
                     end;
               #39,'"' : begin
                           activequote:=lch;
                           inc(pind);
                         end;
               'C' : begin
                       intscandate(@shortdateformat[1],length(ShortDateFormat));
                       intscandate(@longtimeformat[1],length(longtimeformat));
                       inc(pind);
                     end;
               '?' : begin
                       inc(pind);
                       inc(ind);
                     end;
               #9  : begin
                       while (ind<=len) and (s[ind] in whitespace) do
                         inc(ind);
                       inc(pind);
                     end;
               else
                 matchchar(ptrn[pind]);
             end; {case}
             lasttoken:=lch;
            end
          else
            begin
              if activequote=lch then
                begin
                  activequote:=#0;
                  inc(pind);
                end
              else
                matchchar(ptrn[pind]);
            end;
     end;
   if (pind<plen) and (plen>0) and (ptrn[plen-1]<>#9) then  // allow omission of trailing whitespace
     RaiseException('end pattern not matched');
end;

var plen:integer;

begin
  activequote:=#0;
  yy:=0; mm:=0; dd:=0;
  timeval:=0.0;
  len:=length(s); ind:=startpos;
  plen:=length(pattern);
  intscandate(@pattern[1],plen);
  result:=timeval;
  if (yy>0) and (mm>0) and (dd>0) then
     result:=result+encodedate(yy,mm,dd);
end;

procedure dotest(pt,s:string;shouldfail :boolean=false);

var dt : Tdatetime;
    s2 : string;
    fails : boolean;
begin
  fails:=false;
  try
  write(pt:20,s:20,' ':5);
  dt:=scandatetime(pt,s);
  writeln(formatdatetime('yyyy mm dd hh:nn:ss',dt));
  except
    on e : EConvertError do
      begin
        fails:=true;
        writeln(e.message);
      end;
  end;
  if fails<>shouldfail then
   writeln(' BAD!');
end;

var dt:TdateTime;
begin
  writeln('shortdateformat:',shortdateformat);
  writeln('shorttimeformat:',shorttimeformat);
  writeln('longtimeformat :',longtimeformat);
  writeln('shorttimeformat:',shorttimeformat);
  writeln('timeamstring   :',timeamstring);
  writeln('timepmstring   :',timepmstring);


  dotest('hh:nn:ss','10:20:30');
  dotest('yyyy-mm-dd hh:nn:ss','2006-10-06 10:20:30');
  dotest('hhnnss','102030');
  dotest('yyyymmddhhnnss','20061006102030');
  dotest('yymmddhhnnss','061006102030');
  dotest('yy-mm-dd','06-5-3');
  dotest('yy-mmm-dd','06-mar-3');
  dotest('yy-mmmm-dd','06-mar-3',true);    // should fail "mar" is not a long name.
  dotest('mmmm dd?, yyyy','March 4rd, 2006',true);  // should fail one ? too few causes yyyy mismatch
  dotest('mmmm dd??, yyyy','March 4rd, 2006');
  dotest('yy-mm-dd hh:nn:ss ampm','06-5-3 05:23:12 pm');
  dotest('''yy-mm-dd'' "yy-mm-dd" yy-mm-dd hh:nn:ss ampm','yy-mm-dd yy-mm-dd 06-5-3 05:23:12 pm');
  dotest('c','12/06/20076:3:1 PM');
  dotest(#9'yy'#9'mm'#9'dd'#9,'08      10              12');     // also optional trailing whitespace pattern char test
  readln;
end.
